import dataclasses
import inspect
import logging
import os
import unittest
from typing import Any, Dict, Union

import numpy as np

from ..model_config import ModelTypes
from ..model_config.config import get_model_config
from ..model_config.inference_parameter import InferParamManager
from .output import (ClassifierOutput, EmbeddingOutput, ImageOutput, MasksOutput, TextOutput,
                     VisualDetectorOutput)

PREDEFINED_TEXTS = ["Photo of a cat", "A cat is playing around"]

PREDEFINED_IMAGES = [
    np.zeros((100, 100, 3), dtype='uint8'),  #black
    np.ones((100, 100, 3), dtype='uint8') * 255,  #white
    np.random.uniform(0, 255, (100, 100, 3)).astype('uint8')  #noise
]


class DefaultTestInferenceModel(unittest.TestCase):
  """
  This file contains test cases:
  * Test triton config of current model vs default config
  * Test if labels.txt is valid for specific model types
  * Test inference with simple inputs
  ...
  """
  __test__ = False

  def triton_get_predictions(self, input_data, **kwargs):
    """Call InferenceModel.get_predictions method

    Args:
        input_data (Union[np.ndarray, str]):
          if model receives image or vector then type is `np.ndarray`, otherwise `string`

    Returns:
       One of types in models.output
    """
    _kwargs = self.inference_parameters.validate(**kwargs)
    return inspect.unwrap(self.triton_python_model.inference_obj.get_predictions)(
        self.triton_python_model.inference_obj, input_data, **_kwargs)

  def _get_preprocess(self, input):
    """ preprocess if input is image """
    if "image" in input.name:
      h, w, _ = input.dims
      if h > -1 and w > -1:
        import cv2

        def _f(x):
          logging.info(f"Preprocess reshape image => {(w, h, 3)}")
          return cv2.resize(x, (w, h))

        return _f

    return lambda x: x

  def intitialize(self,
                  model_type: str,
                  repo_version_dir: str,
                  is_instance_kind_gpu: bool = True,
                  inference_parameters: Union[str, Dict[str, Any]] = ""):
    import sys
    sys.path.append(repo_version_dir)
    self.model_type = model_type
    self.is_instance_kind_gpu = is_instance_kind_gpu
    logging.info(self.model_type)

    # load inference parameters
    if isinstance(inference_parameters, str):
      self.inference_parameters = InferParamManager(json_path=inference_parameters)
    else:
      self.inference_parameters = InferParamManager.from_kwargs(**inference_parameters)
      exported_file_path = os.path.join(repo_version_dir, "inference_parameters.json")
      logging.info(f"Export inference parameters to `{exported_file_path}` when loading from dict")
      self.inference_parameters.export(exported_file_path)

    # Construct TritonPythonModel object
    from model import TritonPythonModel
    self.triton_python_model = TritonPythonModel()
    self.triton_python_model.initialize(
        dict(
            model_repository=os.path.join(repo_version_dir, ".."),
            model_instance_kind="GPU" if self.is_instance_kind_gpu else "cpu"))
    # Get default config of model and model_type
    self.default_triton_model_config = get_model_config(self.model_type).make_triton_model_config(
        model_name=self.model_type, model_version="1", image_shape=[-1, -1])
    # Get current model config
    self.triton_model_config = self.triton_python_model.config_msg
    self.input_name_to_config = {each.name: each
                                 for each in self.triton_model_config.input}  # name: input
    self.preprocess = {
        k: self._get_preprocess(input)
        for k, input in self.input_name_to_config.items()
    }
    # load labels
    self._required_label_model_types = [
        ModelTypes.visual_detector, ModelTypes.visual_classifier, ModelTypes.text_classifier,
        ModelTypes.visual_segmenter
    ]
    self._output_text_models = [ModelTypes.text_to_text]
    self.labels = []
    if self.model_type in self._required_label_model_types:
      with open(os.path.join(repo_version_dir, "../labels.txt"), 'r') as fp:
        labels = fp.readlines()
        if labels:
          self.labels = [line for line in labels if line]

  def test_triton_config(self):
    """ test Triton config"""
    # check if input names are still matched
    default_input_names = [each.name for each in self.default_triton_model_config.input]
    current_input_names = [each.name for each in self.triton_model_config.input]
    default_input_names.sort()
    current_input_names.sort()
    self.assertEqual(current_input_names, default_input_names,
                     "input name of current model vs generated model must be matched "
                     f"{current_input_names} != {default_input_names}")
    # check if output names are still matched
    default_output_names = [each.name for each in self.default_triton_model_config.output]
    current_output_names = [each.name for each in self.triton_model_config.output]
    default_output_names.sort()
    current_output_names.sort()
    self.assertEqual(current_output_names, default_output_names,
                     "output name of current model vs generated model must be matched "
                     f"{current_output_names} not in {default_output_names}")

  def test_having_labels(self):
    if self.model_type in self._required_label_model_types:
      self.assertTrue(
          len(self.labels),
          f"`labels.txt` is empty!. Model type `{self.model_type}` requires input labels in `labels.txt`"
      )

  def test_inference_with_predefined_inputs(self):
    """ Test Inference with predefined inputs """

    def _is_valid_logit(x: np.array):
      return np.all(0 <= x) and np.all(x <= 1)

    def _is_non_negative(x: np.array):
      return np.all(x >= 0)

    def _is_integer(x):
      return np.all(np.equal(np.mod(x, 1), 0))

    if len(self.input_name_to_config) == 1:
      if "image" in self.preprocess:
        inputs = [self.preprocess["image"](inp) for inp in PREDEFINED_IMAGES]
      else:
        inputs = PREDEFINED_TEXTS
      outputs = [self.triton_get_predictions(inp) for inp in inputs]

      # Test for specific model type:
      # 1. length of output array vs config
      # 2. type of outputs
      # 3. test range value, shape and dtype of output

      for inp, output in zip(inputs, outputs):

        field = dataclasses.fields(output)[0].name
        if self.model_type not in self._output_text_models:
          self.assertEqual(
              len(self.triton_model_config.output[0].dims),
              len(getattr(output, field).shape),
              "Length of 'dims' of config and output must be matched, but get "
              f"Config {len(self.triton_model_config.output[0].dims)} != Output {len(getattr(output, field).shape)}"
          )

        if self.model_type == ModelTypes.visual_detector:
          logging.info(output.predicted_labels)
          self.assertEqual(
              type(output), VisualDetectorOutput,
              f"Output type must be `VisualDetectorOutput`, but got {type(output)}")
          self.assertTrue(
              _is_valid_logit(output.predicted_scores),
              "`predicted_scores` must be in range [0, 1]")
          self.assertTrue(
              _is_non_negative(output.predicted_bboxes), "`predicted_bboxes` must be >= 0")
          self.assertTrue(
              np.all(0 <= output.predicted_labels) and
              np.all(output.predicted_labels < len(self.labels)),
              f"`predicted_labels` must be in [0, {len(self.labels) - 1}]")
          self.assertTrue(
              _is_integer(output.predicted_labels), "`predicted_labels` must be integer")

        elif self.model_type == ModelTypes.visual_classifier:
          self.assertEqual(
              type(output), ClassifierOutput,
              f"Output type must be `ClassifierOutput`, but got {type(output)}")
          self.assertTrue(
              _is_valid_logit(output.predicted_scores),
              "`predicted_scores` must be in range [0, 1]")
          if self.labels:
            self.assertEqual(
                len(output.predicted_scores),
                len(self.labels),
                f"`predicted_labels` must equal to {len(self.labels)}, however got {len(output.predicted_scores)}"
            )

        elif self.model_type == ModelTypes.text_classifier:
          self.assertEqual(
              type(output), ClassifierOutput,
              f"Output type must be `ClassifierOutput`, but got {type(output)}")
          self.assertTrue(
              _is_valid_logit(output.predicted_scores),
              "`predicted_scores` must be in range [0, 1]")
          if self.labels:
            self.assertEqual(
                len(output.predicted_scores),
                len(self.labels),
                f"`predicted_labels` must equal to {len(self.labels)}, however got {len(output.predicted_scores)}"
            )

        elif self.model_type == ModelTypes.text_embedder:
          self.assertEqual(
              type(output), EmbeddingOutput,
              f"Output type must be `EmbeddingOutput`, but got {type(output)}")
          self.assertNotEqual(output.embedding_vector.shape, [])

        elif self.model_type == ModelTypes.text_to_text:
          self.assertEqual(
              type(output), TextOutput,
              f"Output type must be `TextOutput`, but got {type(output)}")

        elif self.model_type == ModelTypes.text_to_image:
          self.assertEqual(
              type(output), ImageOutput,
              f"Output type must be `ImageOutput`, but got {type(output)}")
          self.assertTrue(_is_non_negative(output.image), "`image` elements must be >= 0")

        elif self.model_type == ModelTypes.visual_embedder:
          self.assertEqual(
              type(output), EmbeddingOutput,
              f"Output type must be `EmbeddingOutput`, but got {type(output)}")
          self.assertNotEqual(output.embedding_vector.shape, [])

        elif self.model_type == ModelTypes.visual_segmenter:
          self.assertEqual(
              type(output), MasksOutput,
              f"Output type must be `MasksOutput`, but got {type(output)}")
          self.assertTrue(_is_integer(output.predicted_mask), "`predicted_mask` must be integer")
          if self.labels:
            self.assertTrue(
                np.all(0 <= output.predicted_mask) and
                np.all(output.predicted_mask < len(self.labels)),
                f"`predicted_mask` must be in [0, {len(self.labels) - 1}]")

    elif len(self.input_name_to_config) == 2:
      from itertools import zip_longest
      if self.model_type == ModelTypes.multimodal_embedder:
        input_images = [self.preprocess["image"](inp) for inp in PREDEFINED_IMAGES]
        input_texts = PREDEFINED_TEXTS

        def _assert(input_data):
          for group in zip_longest(*input_data.values()):
            _input = dict(zip(input_data, group))
            output = self.triton_get_predictions(input_data=_input)
            self.assertEqual(
                type(output), EmbeddingOutput,
                f"Output type must be `EmbeddingOutput`, but got {type(output)}")
          self.assertNotEqual(output.embedding_vector.shape, [])

        _assert(dict(image=input_images, text=[]))
        _assert(dict(image=[], text=input_texts))


if __name__ == '__main__':
  unittest.main()



import os
from typing import List

from clarifai.errors import UserError


class Chunker:
  """Split an input sequence into small chunks."""

  def __init__(self, seq: List, size: int) -> None:
    self.seq = seq
    self.size = size

  def chunk(self) -> List[List]:
    """Chunk input sequence."""
    return [self.seq[pos:pos + self.size] for pos in range(0, len(self.seq), self.size)]


class BackoffIterator:
  """Iterator that returns a sequence of backoff values."""

  def __init__(self):
    self.count = 0

  def __iter__(self):
    return self

  def __next__(self):
    if self.count < 1:
      self.count += 1
      return 0.1
    elif self.count < 7:
      self.count += 1
      return 0.01 * (2**(self.count + 4))
    else:
      return 0.01 * (2**10)  # 10 seconds


def get_from_dict_or_env(key: str, env_key: str, **data) -> str:
  """Get a value from a dictionary or an environment variable."""
  if key in data and data[key]:
    return data[key]
  else:
    return get_from_env(key, env_key)


def get_from_env(key: str, env_key: str) -> str:
  """Get a value from a dictionary or an environment variable."""
  if env_key in os.environ and os.environ[env_key]:
    return os.environ[env_key]
  else:
    raise UserError(f"Did not find `{key}`, please add an environment variable"
                    f" `{env_key}` which contains it, or pass"
                    f"  `{key}` as a named parameter.")



# This file contains boilerplate code to allow users write their model
# inference code that will then interact with the Triton Inference Server
# Python backend to serve end user requests.
# The module name, module path, class name & get_predictions() method names MUST be maintained as is
# but other methods may be added within the class as deemed fit provided
# they are invoked within the main get_predictions() inference method
# if they play a role in any step of model inference
"""User model inference script."""

import os
from pathlib import Path

import torch
from transformers import AutoModelForSemanticSegmentation, SegformerImageProcessor

from clarifai.models.model_serving.models.model_types import visual_segmenter
from clarifai.models.model_serving.models.output import MasksOutput


class InferenceModel:
  """User model inference class."""

  def __init__(self) -> None:
    """
    Load inference time artifacts that are called frequently .e.g. models, tokenizers, etc.
    in this method so they are loaded only once for faster inference.
    """
    self.base_path: Path = os.path.dirname(__file__)
    self.huggingface_model_path = os.path.join(self.base_path, "segformer_b2_clothes")
    #self.labels_path = os.path.join(Path(self.base_path).parents[0], "labels.txt")
    self.processor = SegformerImageProcessor.from_pretrained(self.huggingface_model_path)
    self.model = AutoModelForSemanticSegmentation.from_pretrained(self.huggingface_model_path)

  @visual_segmenter
  def get_predictions(self, input_data):
    """
    Main model inference method.

    Args:
    -----
      input_data: A single input data item to predict on.
        Input data can be an image or text, etc depending on the model type.

    Returns:
    --------
      One of the clarifai.models.model_serving.models.output types. Refer to the README/docs
    """
    inputs = self.processor(images=input_data, return_tensors="pt")
    with torch.no_grad():
      output = self.model(**inputs)

    logits = output.logits.cpu()
    mask = logits.argmax(dim=1)[0].numpy()

    return MasksOutput(predicted_mask=mask)



from datetime import datetime
from typing import Any, Callable

from google.protobuf import struct_pb2
from google.protobuf.timestamp_pb2 import Timestamp
from google.protobuf.wrappers_pb2 import BoolValue

from clarifai.client.auth import create_stub
from clarifai.client.auth.helper import ClarifaiAuthHelper
from clarifai.errors import ApiError
from clarifai.utils.misc import get_from_dict_or_env


class BaseClient:
  """BaseClient is the base class for all the classes interacting with Clarifai endpoints.

  Args:
      **kwargs: Additional keyword arguments to be passed to the ClarifaiAuthHelper.
          - user_id (str): A user ID for authentication.
          - app_id (str): An app ID for the application to interact with.
          - pat (str): A personal access token for authentication.
          - base (str): The base URL for the API endpoint. Defaults to 'https://api.clarifai.com'.
          - ui (str): The URL for the UI. Defaults to 'https://clarifai.com'.

  Attributes:
      auth_helper (ClarifaiAuthHelper): An instance of ClarifaiAuthHelper for authentication.
      STUB (Stub): The gRPC Stub object for API interaction.
      metadata (tuple): The gRPC metadata containing the personal access token.
      userDataObject (UserAppIDSet): The protobuf object representing user and app IDs.
      base (str): The base URL for the API endpoint.
  """

  def __init__(self, **kwargs):
    pat = get_from_dict_or_env(key="pat", env_key="CLARIFAI_PAT", **kwargs)
    kwargs.update({'pat': pat})
    self.auth_helper = ClarifaiAuthHelper(**kwargs, validate=False)
    self.STUB = create_stub(self.auth_helper)
    self.metadata = self.auth_helper.metadata
    self.pat = self.auth_helper.pat
    self.user_app_id = self.auth_helper.get_user_app_id_proto()
    self.base = self.auth_helper.base

  def _grpc_request(self, method: Callable, argument: Any):
    """Makes a gRPC request to the API.

    Args:
        method (Callable): The gRPC method to call.
        argument (Any): The argument to pass to the gRPC method.

    Returns:
        res (Any): The result of the gRPC method call.
    """

    try:
      res = method(argument)
      # MessageToDict(res) TODO global debug logger
      return res
    except ApiError:
      raise Exception("ApiError")

  def convert_string_to_timestamp(self, date_str) -> Timestamp:
    """Converts a string to a Timestamp object.

    Args:
        date_str (str): The string to convert.

    Returns:
        Timestamp: The converted Timestamp object.
    """
    # Parse the string into a Python datetime object
    try:
      datetime_obj = datetime.strptime(date_str, '%Y-%m-%dT%H:%M:%S.%fZ')
    except ValueError:
      try:
        datetime_obj = datetime.strptime(date_str, '%Y-%m-%dT%H:%M:%SZ')
      except ValueError:
        return Timestamp()

    # Convert the datetime object to a Timestamp object
    timestamp_obj = Timestamp()
    timestamp_obj.FromDatetime(datetime_obj)

    return timestamp_obj

  def process_response_keys(self, old_dict, listing_resource=None):
    """Converts keys in a response dictionary to resource proto format.

    Args:
        old_dict (dict): The dictionary to convert.

    Returns:
        new_dict (dict): The dictionary with processed keys.
    """
    if listing_resource:
      old_dict[f'{listing_resource}_id'] = old_dict['id']
      old_dict.pop('id')

    def convert_recursive(item):
      if isinstance(item, dict):
        new_item = {}
        for key, value in item.items():
          if key == 'default_value':
            # Map infer param value to proto value
            value_map = dict(number_value=None, string_value=None, bool_value=None)

            def map_fn(v):
              return 'number_value' if isinstance(v, float) or isinstance(v, int) else \
              'string_value' if isinstance(v, str) else \
              'bool_value' if isinstance(v, bool) else None

            value_map[map_fn(value)] = value
            value = struct_pb2.Value(**value_map)
          if key in ['created_at', 'modified_at', 'completed_at']:
            value = self.convert_string_to_timestamp(value)
          elif key in ['workflow_recommended']:
            value = BoolValue(value=True)
          elif key in ['fields_map', 'params']:
            value_s = struct_pb2.Struct()
            value_s.update(value)
            value = value_s
          elif key in ['metadata']:
            continue  # TODO Fix "app_duplication"
          new_item[key] = convert_recursive(value)
        return new_item
      elif isinstance(item, list):
        return [convert_recursive(element) for element in item]
      else:
        return item

    new_dict = convert_recursive(old_dict)
    return new_dict



import logging
import os
import unittest

from clarifai.models.model_serving.models.default_test import DefaultTestInferenceModel


class CustomTestInferenceModel(DefaultTestInferenceModel):
  """
  Run this file to test your implementation of InferenceModel in inference.py with default tests of Triton configuration and its output values based on basic predefined inputs
  If you want to write custom testcase or just test output value.
  Please follow these instrucitons:
  1. Name your test function with prefix "test" so that pytest can execute
  2. In order to obtain output of InferenceModel, call `self.triton_get_predictions(input_data)`.
  3. If your input is `image` and you have set custom size of it when building model repository,
  call `self.preprocess(image)` to obtain correct resized input
  4. Run this test by calling
  ```bash
  pytest ./your_triton_folder/1/test.py
  #to see std output
  pytest --log-cli-level=INFO  -s ./your_triton_folder/1/test.py
  ```

  ### Examples:
  + test text-to-image output
  ```
  def test_text_to_image_output(self):
    text = "Test text"
    output = self.triton_get_predictions(text)
    image = output.image # uint8 np.ndarray image
    #show or save
  ```
  + test visual-classifier output
  ```
  def test_visual_classifier(self):
    image = cv2.imread("your/local/image.jpg") # Keep in mind of format of image (BGR or RGB)
    output = self.triton_get_predictions(image)
    scores = output.predicted_scores # np.ndarray
    #process scores to get class id and its score
    logger.info(result)
  """

  # Insert your inference parameters json path here
  # or insert a dictionary of your_parameter_name and value, e.g dict(x=1.5, y="text", c=True)
  # or Leave it as "" if you don't have it.
  inference_parameters = ""

  ########### Initialization. Do not change it ###########
  __test__ = True

  def setUp(self) -> None:
    logging.info("Initializing...")
    model_type = "clarifai-model-type"  # your model type
    self.intitialize(
        model_type,
        repo_version_dir=os.path.dirname(__file__),
        is_instance_kind_gpu=True,
        inference_parameters=self.inference_parameters)

  ########################################################


if __name__ == '__main__':
  unittest.main()



import os
import time
from typing import Any, Dict, Generator, List

import requests
import yaml
from clarifai_grpc.grpc.api import resources_pb2, service_pb2
from clarifai_grpc.grpc.api.resources_pb2 import Input
from clarifai_grpc.grpc.api.status import status_code_pb2
from google.protobuf.json_format import MessageToDict
from google.protobuf.struct_pb2 import Struct

from clarifai.client.base import BaseClient
from clarifai.client.input import Inputs
from clarifai.client.lister import Lister
from clarifai.constants.model import TRAINABLE_MODEL_TYPES
from clarifai.errors import UserError
from clarifai.urls.helper import ClarifaiUrlHelper
from clarifai.utils.logging import get_logger
from clarifai.utils.misc import BackoffIterator
from clarifai.utils.model_train import (find_and_replace_key, params_parser,
                                        response_to_model_params, response_to_param_info,
                                        response_to_templates)


class Model(Lister, BaseClient):
  """Model is a class that provides access to Clarifai API endpoints related to Model information."""

  def __init__(self,
               url: str = None,
               model_id: str = None,
               model_version: Dict = {'id': ""},
               base_url: str = "https://api.clarifai.com",
               pat: str = None,
               **kwargs):
    """Initializes a Model object.

    Args:
        url (str): The URL to initialize the model object.
        model_id (str): The Model ID to interact with.
        model_version (dict): The Model Version to interact with.
        base_url (str): Base API url. Default "https://api.clarifai.com"
        pat (str): A personal access token for authentication. Can be set as env var CLARIFAI_PAT
        **kwargs: Additional keyword arguments to be passed to the Model.
    """
    if url and model_id:
      raise UserError("You can only specify one of url or model_id.")
    if not url and not model_id:
      raise UserError("You must specify one of url or model_id.")
    if url:
      user_id, app_id, _, model_id, model_version_id = ClarifaiUrlHelper.split_clarifai_url(url)
      model_version = {'id': model_version_id}
      kwargs = {'user_id': user_id, 'app_id': app_id}
    self.kwargs = {**kwargs, 'id': model_id, 'model_version': model_version,}
    self.model_info = resources_pb2.Model(**self.kwargs)
    self.logger = get_logger(logger_level="INFO")
    self.training_params = {}
    BaseClient.__init__(self, user_id=self.user_id, app_id=self.app_id, base=base_url, pat=pat)
    Lister.__init__(self)

  def list_training_templates(self) -> List[str]:
    """Lists all the training templates for the model type.

    Returns:
        templates (List): List of training templates for the model type.

    Example:
        >>> from clarifai.client.model import Model
        >>> model = Model(model_id='model_id', user_id='user_id', app_id='app_id')
        >>> print(model.list_training_templates())
    """
    if not self.model_info.model_type_id:
      self.load_info()
    if self.model_info.model_type_id not in TRAINABLE_MODEL_TYPES:
      raise UserError(f"Model type {self.model_info.model_type_id} is not trainable")
    request = service_pb2.ListModelTypesRequest(user_app_id=self.user_app_id,)
    response = self._grpc_request(self.STUB.ListModelTypes, request)
    if response.status.code != status_code_pb2.SUCCESS:
      raise Exception(response.status)
    templates = response_to_templates(
        response=response, model_type_id=self.model_info.model_type_id)

    return templates

  def get_params(self, template: str = None, save_to: str = 'params.yaml') -> Dict[str, Any]:
    """Returns the model params for the model type and yaml file.

    Args:
        template (str): The template to use for the model type.
        yaml_file (str): The yaml file to save the model params.

    Returns:
        params (Dict): Dictionary of model params for the model type.

    Example:
        >>> from clarifai.client.model import Model
        >>> model = Model(model_id='model_id', user_id='user_id', app_id='app_id')
        >>> model_params = model.get_params(template='template', yaml_file='model_params.yaml')
    """
    if not self.model_info.model_type_id:
      self.load_info()
    if self.model_info.model_type_id not in TRAINABLE_MODEL_TYPES:
      raise UserError(f"Model type {self.model_info.model_type_id} is not trainable")
    if template is None and self.model_info.model_type_id not in [
        "clusterer", "embedding-classifier"
    ]:
      raise UserError(
          f"Template should be provided for {self.model_info.model_type_id} model type")
    if template is not None and self.model_info.model_type_id in [
        "clusterer", "embedding-classifier"
    ]:
      raise UserError(
          f"Template should not be provided for {self.model_info.model_type_id} model type")

    request = service_pb2.ListModelTypesRequest(user_app_id=self.user_app_id,)
    response = self._grpc_request(self.STUB.ListModelTypes, request)
    if response.status.code != status_code_pb2.SUCCESS:
      raise Exception(response.status)
    params = response_to_model_params(
        response=response, model_type_id=self.model_info.model_type_id, template=template)
    #yaml file
    assert save_to.endswith('.yaml'), "File extension should be .yaml"
    with open(save_to, 'w') as f:
      yaml.dump(params, f, default_flow_style=False, sort_keys=False)
    #updating the global model params
    self.training_params.update(params)

    return params

  def update_params(self, **kwargs) -> None:
    """Updates the model params for the model.

    Args:
        **kwargs: model params to update.

    Example:
        >>> from clarifai.client.model import Model
        >>> model = Model(model_id='model_id', user_id='user_id', app_id='app_id')
        >>> model_params = model.get_params(template='template', yaml_file='model_params.yaml')
        >>> model.update_params(batch_size = 8, dataset_version = 'dataset_version_id')
    """
    if self.model_info.model_type_id not in TRAINABLE_MODEL_TYPES:
      raise UserError(f"Model type {self.model_info.model_type_id} is not trainable")
    if len(self.training_params) == 0:
      raise UserError(
          f"Run 'model.get_params' to get the params for the {self.model_info.model_type_id} model type"
      )
    #getting all the keys in nested dictionary
    all_keys = [key for key in self.training_params.keys()] + [
        key for key in self.training_params.values() if isinstance(key, dict) for key in key
    ]
    #checking if the given params are valid
    if not set(kwargs.keys()).issubset(all_keys):
      raise UserError("Invalid params")
    #updating the global model params
    for key, value in kwargs.items():
      find_and_replace_key(self.training_params, key, value)

  def get_param_info(self, param: str) -> Dict[str, Any]:
    """Returns the param info for the param.

    Args:
        param (str): The param to get the info for.

    Returns:
        param_info (Dict): Dictionary of model param info for the param.

    Example:
        >>> from clarifai.client.model import Model
        >>> model = Model(model_id='model_id', user_id='user_id', app_id='app_id')
        >>> model_params = model.get_params(template='template', yaml_file='model_params.yaml')
        >>> model.get_param_info('param')
    """
    if self.model_info.model_type_id not in TRAINABLE_MODEL_TYPES:
      raise UserError(f"Model type {self.model_info.model_type_id} is not trainable")
    if len(self.training_params) == 0:
      raise UserError(
          f"Run 'model.get_params' to get the params for the {self.model_info.model_type_id} model type"
      )

    all_keys = [key for key in self.training_params.keys()] + [
        key for key in self.training_params.values() if isinstance(key, dict) for key in key
    ]
    if param not in all_keys:
      raise UserError(f"Invalid param: '{param}' for model type '{self.model_info.model_type_id}'")
    template = self.training_params['train_params']['template'] if 'template' in all_keys else None

    request = service_pb2.ListModelTypesRequest(user_app_id=self.user_app_id,)
    response = self._grpc_request(self.STUB.ListModelTypes, request)
    if response.status.code != status_code_pb2.SUCCESS:
      raise Exception(response.status)
    param_info = response_to_param_info(
        response=response,
        model_type_id=self.model_info.model_type_id,
        param=param,
        template=template)

    return param_info

  def train(self, yaml_file: str = None) -> str:
    """Trains the model based on the given yaml file or model params.

    Args:
        yaml_file (str): The yaml file for the model params.

    Returns:
        model_version_id (str): The model version ID for the model.

    Example:
        >>> from clarifai.client.model import Model
        >>> model = Model(model_id='model_id', user_id='user_id', app_id='app_id')
        >>> model_params = model.get_params(template='template', yaml_file='model_params.yaml')
        >>> model.train('model_params.yaml')
    """
    if self.model_info.model_type_id not in TRAINABLE_MODEL_TYPES:
      raise UserError(f"Model type {self.model_info.model_type_id} is not trainable")
    if not yaml_file and len(self.training_params) == 0:
      raise UserError("Provide yaml file or run 'model.get_params()'")

    if yaml_file:
      with open(yaml_file, 'r') as file:
        params_dict = yaml.safe_load(file)
    else:
      params_dict = self.training_params

    train_dict = params_parser(params_dict)
    request = service_pb2.PostModelVersionsRequest(
        user_app_id=self.user_app_id,
        model_id=self.id,
        model_versions=[resources_pb2.ModelVersion(**train_dict)])
    response = self._grpc_request(self.STUB.PostModelVersions, request)
    if response.status.code != status_code_pb2.SUCCESS:
      raise Exception(response.status)
    self.logger.info("\nModel Training Started\n%s", response.status)

    return response.model.model_version.id

  def training_status(self, version_id: str, training_logs: bool = False) -> Dict[str, str]:
    """Get the training status for the model version. Also stores training logs

    Args:
        version_id (str): The version ID to get the training status for.
        training_logs (bool): Whether to save the training logs in a file.

    Returns:
        training_status (Dict): Dictionary of training status for the model version.

    Example:
        >>> from clarifai.client.model import Model
        >>> model = Model(model_id='model_id', user_id='user_id', app_id='app_id')
        >>> model.training_status(version_id='version_id',training_logs=True)
    """
    if self.model_info.model_type_id not in TRAINABLE_MODEL_TYPES:
      raise UserError(f"Model type {self.model_info.model_type_id} is not trainable")

    request = service_pb2.GetModelVersionRequest(
        user_app_id=self.user_app_id, model_id=self.id, version_id=version_id)
    response = self._grpc_request(self.STUB.GetModelVersion, request)
    if response.status.code != status_code_pb2.SUCCESS:
      raise Exception(response.status)

    if training_logs:
      try:
        if response.model_version.train_log:
          log_response = requests.get(response.model_version.train_log)
          log_response.raise_for_status()  # Check for any HTTP errors
          with open(version_id + '.log', 'wb') as file:
            for chunk in log_response.iter_content(chunk_size=4096):  # 4KB
              file.write(chunk)
          self.logger.info(f"\nTraining logs are saving in '{version_id+'.log'}' file")

      except requests.exceptions.RequestException as e:
        raise Exception(f"An error occurred while getting training logs: {e}")

    return response.model_version.status

  def delete_version(self, version_id: str) -> None:
    """Deletes a model version for the Model.

    Args:
        version_id (str): The version ID to delete.

    Example:
        >>> from clarifai.client.model import Model
        >>> model = Model(model_id='model_id', user_id='user_id', app_id='app_id')
        >>> model.delete_version(version_id='version_id')
    """
    request = service_pb2.DeleteModelVersionRequest(
        user_app_id=self.user_app_id, model_id=self.id, version_id=version_id)

    response = self._grpc_request(self.STUB.DeleteModelVersion, request)
    if response.status.code != status_code_pb2.SUCCESS:
      raise Exception(response.status)
    self.logger.info("\nModel Version Deleted\n%s", response.status)

  def create_version(self, **kwargs) -> 'Model':
    """Creates a model version for the Model.

    Args:
        **kwargs: Additional keyword arguments to be passed to Model Version.
          - description (str): The description of the model version.
          - concepts (list[Concept]): The concepts to associate with the model version.
          - output_info (resources_pb2.OutputInfo(): The output info to associate with the model version.

    Returns:
        Model: A Model object for the specified model ID.

    Example:
        >>> from clarifai.client.model import Model
        >>> model = Model("url")
                    or
        >>> model = Model(model_id='model_id', user_id='user_id', app_id='app_id')
        >>> model_version = model.create_version(description='model_version_description')
    """
    if self.model_info.model_type_id in TRAINABLE_MODEL_TYPES:
      raise UserError(
          f"{self.model_info.model_type_id} is a trainable model type. Use 'model.train()' to train the model"
      )

    request = service_pb2.PostModelVersionsRequest(
        user_app_id=self.user_app_id,
        model_id=self.id,
        model_versions=[resources_pb2.ModelVersion(**kwargs)])

    response = self._grpc_request(self.STUB.PostModelVersions, request)
    if response.status.code != status_code_pb2.SUCCESS:
      raise Exception(response.status)
    self.logger.info("\nModel Version created\n%s", response.status)

    kwargs.update({'app_id': self.app_id, 'user_id': self.user_id})
    dict_response = MessageToDict(response, preserving_proto_field_name=True)
    kwargs = self.process_response_keys(dict_response['model'], 'model')

    return Model(base_url=self.base, pat=self.pat, **kwargs)

  def list_versions(self, page_no: int = None,
                    per_page: int = None) -> Generator['Model', None, None]:
    """Lists all the versions for the model.

    Args:
        page_no (int): The page number to list.
        per_page (int): The number of items per page.

    Yields:
        Model: Model objects for the versions of the model.

    Example:
        >>> from clarifai.client.model import Model
        >>> model = Model("url") # Example URL: https://clarifai.com/clarifai/main/models/general-image-recognition
                    or
        >>> model = Model(model_id='model_id', user_id='user_id', app_id='app_id')
        >>> all_model_versions = list(model.list_versions())

    Note:
        Defaults to 16 per page if page_no is specified and per_page is not specified.
        If both page_no and per_page are None, then lists all the resources.
    """
    request_data = dict(
        user_app_id=self.user_app_id,
        model_id=self.id,
    )
    all_model_versions_info = self.list_pages_generator(
        self.STUB.ListModelVersions,
        service_pb2.ListModelVersionsRequest,
        request_data,
        per_page=per_page,
        page_no=page_no)

    for model_version_info in all_model_versions_info:
      model_version_info['id'] = model_version_info['model_version_id']
      del model_version_info['model_version_id']
      try:
        del model_version_info['train_info']['dataset']['version']['metrics']
      except KeyError:
        pass
      yield Model(
          model_id=self.id,
          base_url=self.base,
          pat=self.pat,
          **dict(self.kwargs, model_version=model_version_info))

  def predict(self, inputs: List[Input], inference_params: Dict = {}, output_config: Dict = {}):
    """Predicts the model based on the given inputs.

    Args:
        inputs (list[Input]): The inputs to predict, must be less than 128.
    """
    if not isinstance(inputs, list):
      raise UserError('Invalid inputs, inputs must be a list of Input objects.')
    if len(inputs) > 128:
      raise UserError("Too many inputs. Max is 128.")  # TODO Use Chunker for inputs len > 128

    self._override_model_version(inference_params, output_config)
    request = service_pb2.PostModelOutputsRequest(
        user_app_id=self.user_app_id,
        model_id=self.id,
        version_id=self.model_version.id,
        inputs=inputs,
        model=self.model_info)

    start_time = time.time()
    backoff_iterator = BackoffIterator()
    while True:
      response = self._grpc_request(self.STUB.PostModelOutputs, request)

      if response.status.code == status_code_pb2.MODEL_DEPLOYING and \
        time.time() - start_time < 60 * 10: # 10 minutes
        self.logger.info(f"{self.id} model is still deploying, please wait...")
        time.sleep(next(backoff_iterator))
        continue

      if response.status.code != status_code_pb2.SUCCESS:
        raise Exception(f"Model Predict failed with response {response.status!r}")
      else:
        break

    return response

  def predict_by_filepath(self,
                          filepath: str,
                          input_type: str,
                          inference_params: Dict = {},
                          output_config: Dict = {}):
    """Predicts the model based on the given filepath.

    Args:
        filepath (str): The filepath to predict.
        input_type (str): The type of input. Can be 'image', 'text', 'video' or 'audio.
        inference_params (dict): The inference params to override.
        output_config (dict): The output config to override.
          min_value (float): The minimum value of the prediction confidence to filter.
          max_concepts (int): The maximum number of concepts to return.
          select_concepts (list[Concept]): The concepts to select.

    Example:
        >>> from clarifai.client.model import Model
        >>> model = Model("url") # Example URL: https://clarifai.com/clarifai/main/models/general-image-recognition
                    or
        >>> model = Model(model_id='model_id', user_id='user_id', app_id='app_id')
        >>> model_prediction = model.predict_by_filepath('/path/to/image.jpg', 'image')
        >>> model_prediction = model.predict_by_filepath('/path/to/text.txt', 'text')
    """
    if not os.path.isfile(filepath):
      raise UserError('Invalid filepath.')

    with open(filepath, "rb") as f:
      file_bytes = f.read()

    return self.predict_by_bytes(file_bytes, input_type, inference_params, output_config)

  def predict_by_bytes(self,
                       input_bytes: bytes,
                       input_type: str,
                       inference_params: Dict = {},
                       output_config: Dict = {}):
    """Predicts the model based on the given bytes.

    Args:
        input_bytes (bytes): File Bytes to predict on.
        input_type (str): The type of input. Can be 'image', 'text', 'video' or 'audio.
        inference_params (dict): The inference params to override.
        output_config (dict): The output config to override.
          min_value (float): The minimum value of the prediction confidence to filter.
          max_concepts (int): The maximum number of concepts to return.
          select_concepts (list[Concept]): The concepts to select.

    Example:
        >>> from clarifai.client.model import Model
        >>> model = Model("https://clarifai.com/openai/chat-completion/models/GPT-4")
        >>> model_prediction = model.predict_by_bytes(b'Write a tweet on future of AI',
                                                      input_type='text',
                                                      inference_params=dict(temperature=str(0.7), max_tokens=30)))
    """
    if input_type not in {'image', 'text', 'video', 'audio'}:
      raise UserError(
          f"Got input type {input_type} but expected one of image, text, video, audio.")
    if not isinstance(input_bytes, bytes):
      raise UserError('Invalid bytes.')

    if input_type == "image":
      input_proto = Inputs.get_input_from_bytes("", image_bytes=input_bytes)
    elif input_type == "text":
      input_proto = Inputs.get_input_from_bytes("", text_bytes=input_bytes)
    elif input_type == "video":
      input_proto = Inputs.get_input_from_bytes("", video_bytes=input_bytes)
    elif input_type == "audio":
      input_proto = Inputs.get_input_from_bytes("", audio_bytes=input_bytes)

    return self.predict(
        inputs=[input_proto], inference_params=inference_params, output_config=output_config)

  def predict_by_url(self,
                     url: str,
                     input_type: str,
                     inference_params: Dict = {},
                     output_config: Dict = {}):
    """Predicts the model based on the given URL.

    Args:
        url (str): The URL to predict.
        input_type (str): The type of input. Can be 'image', 'text', 'video' or 'audio.
        inference_params (dict): The inference params to override.
        output_config (dict): The output config to override.
          min_value (float): The minimum value of the prediction confidence to filter.
          max_concepts (int): The maximum number of concepts to return.
          select_concepts (list[Concept]): The concepts to select.

    Example:
        >>> from clarifai.client.model import Model
        >>> model = Model("url") # Example URL: https://clarifai.com/clarifai/main/models/general-image-recognition
                    or
        >>> model = Model(model_id='model_id', user_id='user_id', app_id='app_id')
        >>> model_prediction = model.predict_by_url('url', 'image')
    """
    if input_type not in {'image', 'text', 'video', 'audio'}:
      raise UserError(
          f"Got input type {input_type} but expected one of image, text, video, audio.")

    if input_type == "image":
      input_proto = Inputs.get_input_from_url("", image_url=url)
    elif input_type == "text":
      input_proto = Inputs.get_input_from_url("", text_url=url)
    elif input_type == "video":
      input_proto = Inputs.get_input_from_url("", video_url=url)
    elif input_type == "audio":
      input_proto = Inputs.get_input_from_url("", audio_url=url)

    return self.predict(
        inputs=[input_proto], inference_params=inference_params, output_config=output_config)

  def _override_model_version(self, inference_params: Dict = {}, output_config: Dict = {}) -> None:
    """Overrides the model version.

    Args:
        inference_params (dict): The inference params to override.
        output_config (dict): The output config to override.
          min_value (float): The minimum value of the prediction confidence to filter.
          max_concepts (int): The maximum number of concepts to return.
          select_concepts (list[Concept]): The concepts to select.
          sample_ms (int): The number of milliseconds to sample.
    """
    params = Struct()
    if inference_params is not None:
      params.update(inference_params)

    self.model_info.model_version.output_info.CopyFrom(
        resources_pb2.OutputInfo(
            output_config=resources_pb2.OutputConfig(**output_config), params=params))

  def load_info(self) -> None:
    """Loads the model info."""
    request = service_pb2.GetModelRequest(
        user_app_id=self.user_app_id,
        model_id=self.id,
        version_id=self.model_info.model_version.id)
    response = self._grpc_request(self.STUB.GetModel, request)

    if response.status.code != status_code_pb2.SUCCESS:
      raise Exception(response.status)

    dict_response = MessageToDict(response, preserving_proto_field_name=True)
    self.kwargs = self.process_response_keys(dict_response['model'])
    self.model_info = resources_pb2.Model(**self.kwargs)

  def __getattr__(self, name):
    return getattr(self.model_info, name)

  def __str__(self):
    if len(self.kwargs) < 10:
      self.load_info()

    init_params = [param for param in self.kwargs.keys()]
    attribute_strings = [
        f"{param}={getattr(self.model_info, param)}" for param in init_params
        if hasattr(self.model_info, param)
    ]
    return f"Model Details: \n{', '.join(attribute_strings)}\n"



import os

import pytest
from clarifai_grpc.grpc.api import resources_pb2

from clarifai.client.input import Inputs
from clarifai.client.model import Model
from clarifai.errors import UserError

DOG_IMAGE_URL = "https://samples.clarifai.com/dog2.jpeg"
NON_EXISTING_IMAGE_URL = "http://example.com/non-existing.jpg"
RED_TRUCK_IMAGE_FILE_PATH = os.path.dirname(__file__) + "/assets/red-truck.png"
BEER_VIDEO_URL = "https://samples.clarifai.com/beer.mp4"

MAIN_APP_ID = "main"
MAIN_APP_USER_ID = "clarifai"
GENERAL_MODEL_ID = "aaa03c23b3724a16a56b629203edc62c"
CLIP_EMBED_MODEL_ID = "multimodal-clip-embed"

RAW_TEXT = "Hi my name is Jim."
RAW_TEXT_BYTES = b"Hi my name is Jim."

CLARIFAI_PAT = os.environ["CLARIFAI_PAT"]


@pytest.fixture
def model():
  return Model(
      user_id=MAIN_APP_USER_ID, app_id=MAIN_APP_ID, model_id=GENERAL_MODEL_ID, pat=CLARIFAI_PAT)


def validate_concepts_length(response):
  assert len(response.outputs[0].data.concepts) > 0


def test_predict_image_url(model):
  response = model.predict_by_url(DOG_IMAGE_URL, 'image')
  validate_concepts_length(response)


def test_predict_filepath(model):
  response = model.predict_by_filepath(RED_TRUCK_IMAGE_FILE_PATH, 'image')
  validate_concepts_length(response)


def test_predict_image_bytes(model):
  with open(RED_TRUCK_IMAGE_FILE_PATH, "rb") as f:
    image_bytes = f.read()

  response = model.predict_by_bytes(image_bytes, 'image')
  validate_concepts_length(response)


def test_predict_image_url_with_selected_concepts():
  selected_concepts = [
      resources_pb2.Concept(name="dog"),
      resources_pb2.Concept(name="cat"),
  ]
  model_with_selected_concepts = Model(
      user_id=MAIN_APP_USER_ID, app_id=MAIN_APP_ID, model_id=GENERAL_MODEL_ID)

  response = model_with_selected_concepts.predict_by_url(
      DOG_IMAGE_URL, 'image', output_config=dict(select_concepts=selected_concepts))
  concepts = response.outputs[0].data.concepts

  assert len(concepts) == 2
  dog_concept = next(c for c in concepts if c.name == "dog")
  cat_concept = next(c for c in concepts if c.name == "cat")
  assert dog_concept.value > cat_concept.value


def test_predict_image_url_with_min_value():
  model_with_min_value = Model(
      user_id=MAIN_APP_USER_ID,
      app_id=MAIN_APP_ID,
      model_id=GENERAL_MODEL_ID,
  )

  response = model_with_min_value.predict_by_url(
      DOG_IMAGE_URL, 'image', output_config=dict(min_value=0.98))
  assert len(response.outputs[0].data.concepts) > 0
  for c in response.outputs[0].data.concepts:
    assert c.value >= 0.98


def test_predict_image_url_with_max_concepts():
  model_with_max_concepts = Model(
      user_id=MAIN_APP_USER_ID, app_id=MAIN_APP_ID, model_id=GENERAL_MODEL_ID)

  response = model_with_max_concepts.predict_by_url(
      DOG_IMAGE_URL, 'image', output_config=dict(max_concepts=3))
  assert len(response.outputs[0].data.concepts) == 3


def test_failed_predicts(model):
  # Invalid FilePath
  false_filepath = "false_filepath"
  with pytest.raises(UserError):
    model.predict_by_filepath(false_filepath, 'image')

  # Invalid URL
  with pytest.raises(Exception):
    model.predict_by_url(NON_EXISTING_IMAGE_URL, 'image')

  # Invalid Input Type
  with pytest.raises(UserError):
    model.predict_by_url(DOG_IMAGE_URL, 'invalid_input_type')


def test_predict_video_url_with_custom_sample_ms():
  model_with_custom_sample_ms = Model(
      user_id=MAIN_APP_USER_ID,
      app_id=MAIN_APP_ID,
      model_id=GENERAL_MODEL_ID,
  )
  video_proto = Inputs.get_input_from_url("", video_url=BEER_VIDEO_URL)
  response = model_with_custom_sample_ms.predict([video_proto], output_config=dict(sample_ms=2000))
  # The expected time per frame is the middle between the start and the end of the frame
  # (in milliseconds).
  expected_time = 1000

  assert len(response.outputs[0].data.frames) > 0
  for frame in response.outputs[0].data.frames:
    assert frame.frame_info.time == expected_time
    expected_time += 2000


def test_text_embed_predict_with_raw_text():
  clip_dim = 512
  clip_embed_model = Model(
      user_id=MAIN_APP_USER_ID, app_id=MAIN_APP_ID, model_id=CLIP_EMBED_MODEL_ID)

  input_text_proto = Inputs.get_input_from_bytes("", text_bytes=RAW_TEXT.encode(encoding='UTF-8'))
  response = clip_embed_model.predict([input_text_proto])
  assert response.outputs[0].data.embeddings[0].num_dimensions == clip_dim

  response = clip_embed_model.predict([input_text_proto])
  assert response.outputs[0].data.embeddings[0].num_dimensions == clip_dim


def test_model_load_info():
  clip_embed_model = Model(
      user_id=MAIN_APP_USER_ID, app_id=MAIN_APP_ID, model_id=CLIP_EMBED_MODEL_ID)
  assert len(clip_embed_model.kwargs) == 4
  clip_embed_model.load_info()
  assert len(clip_embed_model.kwargs) > 10



# Copyright 2023 Clarifai, Inc.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
""" Model Config classes."""

from dataclasses import asdict, dataclass, field
from typing import List

import yaml

from ..models.model_types import *  # noqa # pylint: disable=unused-import
from ..models.output import *  # noqa # pylint: disable=unused-import

__all__ = ["get_model_config", "MODEL_TYPES", "TritonModelConfig", "ModelTypes"]

### Triton Model Config classes.###


@dataclass
class DType:
  """
  Triton Model Config data types.
  """
  # https://github.com/triton-inference-server/common/blob/main/protobuf/model_config.proto
  TYPE_UINT8: int = 2
  TYPE_INT8: int = 6
  TYPE_INT16: int = 7
  TYPE_INT32: int = 8
  TYPE_INT64: int = 9
  TYPE_FP16: int = 10
  TYPE_FP32: int = 11
  TYPE_STRING: int = 13
  KIND_GPU: int = 1
  KIND_CPU: int = 2


@dataclass
class InputConfig:
  """
  Triton Input definition.
  Params:
  -------
  name: input name
  data_type: input data type
  dims: Pre-defined input data shape(s).

  Returns:
  --------
  InputConfig
  """
  name: str
  data_type: int
  dims: List = field(default_factory=list)
  optional: bool = False


@dataclass
class OutputConfig:
  """
  Triton Output definition.
  Params:
  -------
  name: output name
  data_type: output data type
  dims: Pre-defined output data shape(s).
  labels (bool): If labels file is required for inference.

  Returns:
  --------
  OutputConfig
  """
  name: str
  data_type: int
  dims: List = field(default_factory=list)
  labels: bool = False

  def __post_init__(self):
    if self.labels:
      self.label_filename = "labels.txt"
    del self.labels


@dataclass
class Device:
  """
  Triton instance_group.
  Define the type of inference device and number of devices to use.
  Params:
  -------
  count: number of devices
  use_gpu: whether to use cpu or gpu.

  Returns:
  --------
  Device object
  """
  count: int = 1
  use_gpu: bool = True

  def __post_init__(self):
    if self.use_gpu:
      self.kind: str = DType.KIND_GPU
    else:
      self.kind: str = DType.KIND_CPU


@dataclass
class DynamicBatching:
  """
  Triton dynamic_batching config.
  Params:
  -------
  preferred_batch_size: batch size
  max_queue_delay_microseconds: max queue delay for a request batch

  Returns:
  --------
  DynamicBatching object
  """
  #preferred_batch_size: List[int] = [1] # recommended not to set
  max_queue_delay_microseconds: int = 500


@dataclass
class TritonModelConfig:
  """
  Triton Model Config base.
  Params:
  -------
  name: triton inference model name
  input: a list of an InputConfig field
  output: a list of OutputConfig fields/dicts
  instance_group: Device. see Device
  dynamic_batching: Triton dynamic batching settings.
  max_batch_size: max request batch size
  backend: Triton Python Backend. Constant

  Returns:
  --------
  TritonModelConfig
  """
  model_type: str
  model_name: str
  model_version: str
  image_shape: List  #(H, W)
  input: List[InputConfig] = field(default_factory=list)
  output: List[OutputConfig] = field(default_factory=list)
  instance_group: Device = field(default_factory=Device)
  dynamic_batching: DynamicBatching = field(default_factory=DynamicBatching)
  max_batch_size: int = 1
  backend: str = "python"

  def __post_init__(self):
    if "image" in [each.name for each in self.input]:
      image_dims = self.image_shape
      image_dims.append(3)  # add channel dim
      self.input[0].dims = image_dims


### General Model Config classes & functions ###


# Clarifai model types
@dataclass
class ModelTypes:
  visual_detector: str = "visual-detector"
  visual_classifier: str = "visual-classifier"
  text_classifier: str = "text-classifier"
  text_to_text: str = "text-to-text"
  text_embedder: str = "text-embedder"
  text_to_image: str = "text-to-image"
  visual_embedder: str = "visual-embedder"
  visual_segmenter: str = "visual-segmenter"
  multimodal_embedder: str = "multimodal-embedder"

  def __post_init__(self):
    self.all = list(asdict(self).values())


@dataclass
class InferenceConfig:
  wrap_func: callable
  return_type: dataclass


@dataclass
class FieldMapsConfig:
  input_fields_map: dict
  output_fields_map: dict


@dataclass
class DefaultTritonConfig:
  input: List[InputConfig] = field(default_factory=list)
  output: List[OutputConfig] = field(default_factory=list)


@dataclass
class ModelConfigClass:
  type: str = field(init=False)
  triton: DefaultTritonConfig
  inference: InferenceConfig
  field_maps: FieldMapsConfig

  def make_triton_model_config(
      self,
      model_name: str,
      model_version: str,
      image_shape: List = None,
      instance_group: Device = Device(),
      dynamic_batching: DynamicBatching = DynamicBatching(),
      max_batch_size: int = 1,
      backend: str = "python",
  ) -> TritonModelConfig:

    return TritonModelConfig(
        model_type=self.type,
        model_name=model_name,
        model_version=model_version,
        image_shape=image_shape,
        instance_group=instance_group,
        dynamic_batching=dynamic_batching,
        max_batch_size=max_batch_size,
        backend=backend,
        input=self.triton.input,
        output=self.triton.output)


def read_config(cfg: str):
  with open(cfg, encoding="utf-8") as f:
    config = yaml.safe_load(f)  # model dict

  # parse default triton
  input_triton_configs = config["triton"]["input"]
  output_triton_configs = config["triton"]["output"]
  triton = DefaultTritonConfig(
      input=[
          InputConfig(
              name=input["name"],
              data_type=eval(f"DType.{input['data_type']}"),
              dims=input["dims"],
              optional=input.get("optional", False),
          ) for input in input_triton_configs
      ],
      output=[
          OutputConfig(
              name=output["name"],
              data_type=eval(f"DType.{output['data_type']}"),
              dims=output["dims"],
              labels=output["labels"],
          ) for output in output_triton_configs
      ])

  # parse inference config
  inference = InferenceConfig(
      wrap_func=eval(config["inference"]["wrap_func"]),
      return_type=eval(config["inference"]["return_type"]),
  )

  # parse field maps for deployment
  field_maps = FieldMapsConfig(**config["field_maps"])

  return ModelConfigClass(triton=triton, inference=inference, field_maps=field_maps)


def get_model_config(model_type: str) -> ModelConfigClass:
  """
  Get model config by model type

  Args:

    model_type (str): One of field value of ModelTypes

  Return:
    ModelConfigClass

  ### Example:
  >>> cfg = get_model_config(ModelTypes.text_classifier)
  >>> custom_triton_config = cfg.make_triton_model_config(**kwargs)


  """
  import os
  assert model_type in MODEL_TYPES, f"`model_type` must be in {MODEL_TYPES}"
  cfg = read_config(
      os.path.join(os.path.dirname(__file__), "model_types_config", f"{model_type}.yaml"))
  cfg.type = model_type
  return cfg


_model_types = ModelTypes()
MODEL_TYPES = _model_types.all
del _model_types



# Copyright 2023 Clarifai, Inc.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Interface to Clarifai Runners API."""

import os

from clarifai_grpc.grpc.api import resources_pb2, service_pb2
from clarifai_grpc.grpc.api.status import status_code_pb2, status_pb2
from google.protobuf import json_format

from clarifai.client.base import BaseClient
from clarifai.errors import UserError
from clarifai.utils.logging import get_logger


class Runner(BaseClient):
  """Base class for remote inference runners. This should be subclassed with the run_input method
  implemented to process each input in the request.

  Then on the subclass call start() to start the run loop.
  """

  def __init__(self,
               runner_id: str,
               user_id: str = None,
               check_runner_exists: bool = True,
               base_url: str = "https://api.clarifai.com",
               pat: str = None,
               **kwargs) -> None:
    """
    Args:
      runner_id (str): the id of the runner to use. Create the runner in the Clarifai API first
      user_id (str): Clarifai User ID
      base_url (str): Base API url. Default "https://api.clarifai.com"
      pat (str): A personal access token for authentication. Can be set as env var CLARIFAI_PAT
    """
    user_id = user_id or os.environ.get("CLARIFAI_USER_ID", "")

    if not user_id:
      raise UserError(
          "Set CLARIFAI_USER_ID as environment variables or pass user_id as input arguments")

    self.runner_id = runner_id
    self.logger = get_logger("INFO", __name__)
    self.kwargs = {**kwargs, 'id': runner_id, 'user_id': user_id}
    self.runner_info = resources_pb2.Runner(**self.kwargs)
    BaseClient.__init__(self, user_id=self.user_id, app_id="", base=base_url, pat=pat)

    # Check that the runner exists.
    if check_runner_exists:
      request = service_pb2.GetRunnerRequest(user_app_id=self.user_app_id, runner_id=runner_id)
      response = self._grpc_request(self.STUB.GetRunner, request)
      if response.status.code != status_code_pb2.SUCCESS:
        raise Exception(
            f"""Error getting runner, are you use this is a valid runner id {runner_id} at the user_id
            {self.user_app_id.user_id}.
            Error: {response.status.description}""")

  def start(self):
    """Start the run loop. This will ask the Clarifai API for work, and when it gets work, it will run
    the model on the inputs and post the results back to the Clarifai API. It will then ask for more
    work again.
    """
    self._long_poll_loop()

  def _run(self, request: service_pb2.PostModelOutputsRequest) -> service_pb2.MultiOutputResponse:
    """Run the model on the given request. You shouldn't need to override this method, see run_input
    for the implementation to process each input in the request.

    Args:
      request: service_pb2.PostModelOutputsRequest - the request to run the model on

    Returns:
      service_pb2.MultiOutputResponse - the response from the model's run_input implementation.
    """
    outputs = []
    # TODO: parallelize this
    for inp in request.inputs:
      # TODO: handle errors
      outputs.append(self.run_input(inp))

    return service_pb2.MultiOutputResponse(
        status=status_pb2.Status(
            code=status_code_pb2.SUCCESS,
            description="Success",
        ),
        outputs=outputs,
    )

  def run_input(self, input: resources_pb2.Input) -> resources_pb2.Output:
    """Run the model on the given input in the request. This is the method you should override to
    process each input in the request.

    Args:
      input: resources_pb2.Input - the input to run the model on

    Returns:
      resources_pb2.Output - the response from the model's run_input implementation.
    """
    raise NotImplementedError("run_input() not implemented")

  def _long_poll_loop(self):
    """This method will long poll for work, and when it gets work, it will run the model on the inputs
    and post the results back to the Clarifai API. It will then long poll again for more work.
    """
    c = 0
    # TODO: handle more errors within this loop so it never stops.
    # TODO: perhaps have multiple processes running this loop to handle more work.
    while True:
      # Long poll waiting for work.
      self.logger.info("Loop iteration: {}".format(c))
      request = service_pb2.ListRunnerItemsRequest(
          user_app_id=self.user_app_id, runner_id=self.runner_id)
      work_response = self._grpc_request(self.STUB.ListRunnerItems, request)
      if work_response.status.code == status_code_pb2.RUNNER_NEEDS_RETRY:
        c += 1
        continue  # immediate restart the long poll
      if work_response.status.code != status_code_pb2.SUCCESS:
        raise Exception("Error getting work: {}".format(work_response.status.description))
      if len(work_response.items) == 0:
        self.logger.info("No work to do. Waiting...")
        continue

      # We have work to do. Run the model on the inputs.
      for item in work_response.items:
        if not item.HasField('post_model_outputs_request'):
          raise Exception("Unexpected work item type: {}".format(item))
        self.logger.info(
            f"Working on item: {item.id} with inputs {len(item.post_model_outputs_request.inputs)}"
        )
        result = self._run(item.post_model_outputs_request)

        request = service_pb2.PostRunnerItemOutputsRequest(
            user_app_id=self.user_app_id,
            item_id=item.id,
            runner_id=self.runner_id,
            runner_item_outputs=[service_pb2.RunnerItemOutput(multi_output_response=result)])
        result_response = self._grpc_request(self.STUB.PostRunnerItemOutputs, request)
        if result_response.status.code != status_code_pb2.SUCCESS:
          raise Exception(
              json_format.MessageToJson(result_response, preserving_proto_field_name=True))
          # raise Exception("Error posting result: {}".format(result_response.status.description))

  def __getattr__(self, name):
    return getattr(self.runner_info, name)

  def __str__(self):
    init_params = [param for param in self.kwargs.keys()]
    attribute_strings = [
        f"{param}={getattr(self.runner_info, param)}" for param in init_params
        if hasattr(self.runner_info, param)
    ]
    return f"Runner Details: \n{', '.join(attribute_strings)}\n"



#! COCO 2017 Image Segmentation dataset

import gc
import os
from functools import reduce

import cv2
import numpy as np
from pycocotools import mask as maskUtils
from pycocotools.coco import COCO

from clarifai.datasets.upload.base import ClarifaiDataLoader

from ..features import VisualSegmentationFeatures


class COCOSegmentationDataLoader(ClarifaiDataLoader):
  """COCO Image Segmentation Dataset."""

  def __init__(self, images_dir, label_filepath):
    """
    Args:
      images_dir: Directory containing the images.
      label_filepath: Path to the COCO annotation file.
    """
    self.images_dir = images_dir
    self.label_filepath = label_filepath

    self.map_ids = {}
    self.load_data()

  @property
  def task(self):
    return "visual_segmentation"

  def load_data(self) -> None:
    self.coco = COCO(self.label_filepath)
    self.map_ids = {i: img_id for i, img_id in enumerate(list(self.coco.imgs.keys()))}

  def __len__(self):
    return len(self.coco.imgs)

  def __getitem__(self, index):
    """Get image and annotations for a given index."""
    value = self.coco.imgs[self.map_ids[index]]
    image_path = os.path.join(self.images_dir, value['file_name'])
    annots = []  # polygons
    concept_ids = []

    input_ann_ids = self.coco.getAnnIds(imgIds=[value['id']])
    input_anns = self.coco.loadAnns(input_ann_ids)

    for ann in input_anns:
      # get concept info
      # note1: concept_name can be human readable
      # note2: concept_id can only be alphanumeric, up to 32 characters, with no special chars except `-` and `_`
      concept_name = self.coco.cats[ann['category_id']]['name']
      concept_id = concept_name.lower().replace(' ', '-')

      # get polygons
      if isinstance(ann['segmentation'], list):
        poly = np.array(ann['segmentation']).reshape((int(len(ann['segmentation'][0]) / 2),
                                                      2)).astype(float)
        poly[:, 0], poly[:, 1] = poly[:, 0] / value['width'], poly[:, 1] / value['height']
        poly = np.clip(poly, 0, 1)
        annots.append(poly.tolist())  #[[x=col, y=row],...]
        concept_ids.append(concept_id)
      else:  # seg: {"counts":[...]}
        if isinstance(ann['segmentation']['counts'], list):
          rle = maskUtils.frPyObjects([ann['segmentation']], value['height'], value['width'])
        else:
          rle = ann['segmentation']
        mask = maskUtils.decode(rle)  #binary mask
        #convert mask to polygons and add to annots
        contours, _ = cv2.findContours(mask, cv2.RETR_TREE, cv2.CHAIN_APPROX_SIMPLE)
        polygons = []
        for cont in contours:
          if cont.size >= 6:
            polygons.append(cont.astype(float).flatten().tolist())
        # store polygons in (x,y) pairs
        polygons_flattened = reduce(lambda x, y: x + y, polygons)
        del polygons
        del contours
        del mask
        gc.collect()

        polygons = np.array(polygons_flattened).reshape((int(len(polygons_flattened) / 2),
                                                         2)).astype(float)
        polygons[:, 0] = polygons[:, 0] / value['width']
        polygons[:, 1] = polygons[:, 1] / value['height']
        polygons = np.clip(polygons, 0, 1)
        annots.append(polygons.tolist())  #[[x=col, y=row],...,[x=col, y=row]]
        concept_ids.append(concept_id)

    assert len(concept_ids) == len(annots), f"Num concepts must match num bbox annotations\
    for a single image. Found {len(concept_ids)} concepts and {len(annots)} bboxes."

    return VisualSegmentationFeatures(image_path, concept_ids, annots, id=str(value['id']))



# Copyright 2023 Clarifai, Inc.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Triton model repository generation commandline interface."""

import argparse

from ..constants import MAX_HW_DIM
from ..model_config import MODEL_TYPES, get_model_config
from ..pb_model_repository import TritonModelRepository


def dims_type(shape_string: str):
  """Read list string from cli and convert values to a list of integers."""
  shape_string = shape_string.replace("[", "").replace("]", "")
  shapes = list(map(int, shape_string.split(",")))
  return shapes


def model_upload_init():
  """
  Clarifai triton model upload commandline tool.
  """
  parser = argparse.ArgumentParser(description=__doc__)
  # TritonModelConfig args
  parser.add_argument("--model_name", type=str, required=True, help="Inference Model Name")
  parser.add_argument(
      "--model_version",
      type=str,
      default="1",
      required=False,
      help="Triton inference model version name. 1 stands for version 1. \
    Leave as default value (Recommended).")
  parser.add_argument(
      "--model_type",
      type=str,
      choices=MODEL_TYPES,
      required=True,
      help=f"Clarifai supported model types.\n Model-types-map: {MODEL_TYPES}",
  )
  parser.add_argument(
      "--image_shape",
      type=dims_type,
      default="[-1, -1]",
      required=False,
      help="(H, W) dims for models with an image input type. H and W each have a max value of 1024",
  )
  parser.add_argument(
      "--repo_dir",
      type=str,
      default=".",
      required=True,
      help="Directory to create triton repository.")

  args = parser.parse_args()

  if len(args.image_shape) != 2:
    raise ValueError(
        f"image_shape takes 2 values, Height and Width. Got {len(args.image_shape)} values instead."
    )

  if args.image_shape[0] > MAX_HW_DIM or args.image_shape[1] > MAX_HW_DIM:
    raise ValueError(
        f"H and W each have a maximum value of 1024. Got H: {args.image_shape[0]}, W: {args.image_shape[1]}"
    )

  model_config = get_model_config(args.model_type).make_triton_model_config(
      model_name=args.model_name,
      model_version="1",
      image_shape=args.image_shape,
  )

  triton_repo = TritonModelRepository(model_config)
  triton_repo.build_repository(args.repo_dir)


if __name__ == "__main__":
  model_upload_init()
